@@ -15,7 +15,7 @@
  *
  *
  * IDENTIFICATION
- *		$PostgreSQL: pgsql/src/bin/pg_dump/pg_backup_archiver.c,v 1.101.4.4 2005/03/18 17:33:03 tgl Exp $
+ *		$PostgreSQL: pgsql/src/bin/pg_dump/pg_backup_archiver.c,v 1.101.4.9 2006/02/05 20:59:06 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -73,6 +73,8 @@ static void _die_horribly(ArchiveHandle *AH, const char *modulename, const char
 static int	_canRestoreBlobs(ArchiveHandle *AH);
 static int	_restoringToDB(ArchiveHandle *AH);
 
+static void dumpTimestamp(ArchiveHandle *AH, const char *msg, time_t tim);
+
 
 /*
  *	Wrapper functions.
@@ -129,7 +131,7 @@ void
 RestoreArchive(Archive *AHX, RestoreOptions *ropt)
 {
 	ArchiveHandle *AH = (ArchiveHandle *) AHX;
-	TocEntry   *te = AH->toc->next;
+	TocEntry   *te;
 	teReqs		reqs;
 	OutputContext sav;
 	bool		defnDumped;
@@ -210,6 +212,9 @@ RestoreArchive(Archive *AHX, RestoreOptions *ropt)
 
 	ahprintf(AH, "--\n-- PostgreSQL database dump\n--\n\n");
 
+	if (AH->public.verbose)
+		dumpTimestamp(AH, "Started on", AH->createDate);
+
 	/*
 	 * Establish important parameter values right away.
 	 */
@@ -222,11 +227,10 @@ RestoreArchive(Archive *AHX, RestoreOptions *ropt)
 	 */
 	if (ropt->dropSchema)
 	{
-		te = AH->toc->prev;
-		AH->currentTE = te;
-
-		while (te != AH->toc)
+		for (te = AH->toc->prev; te != AH->toc; te = te->prev)
 		{
+			AH->currentTE = te;
+
 			reqs = _tocEntryRequired(te, ropt, false /* needn't drop ACLs */);
 			if (((reqs & REQ_SCHEMA) != 0) && te->dropStmt)
 			{
@@ -238,15 +242,13 @@ RestoreArchive(Archive *AHX, RestoreOptions *ropt)
 				/* Drop it */
 				ahprintf(AH, "%s", te->dropStmt);
 			}
-			te = te->prev;
 		}
 	}
 
 	/*
 	 * Now process each non-ACL TOC entry
 	 */
-	te = AH->toc->next;
-	while (te != AH->toc)
+	for (te = AH->toc->next; te != AH->toc; te = te->next)
 	{
 		AH->currentTE = te;
 
@@ -343,10 +345,15 @@ RestoreArchive(Archive *AHX, RestoreOptions *ropt)
 						 * mode with libpq.
 						 */
 						if (te->copyStmt && strlen(te->copyStmt) > 0)
-							ahprintf(AH, te->copyStmt);
+						{
+							ahprintf(AH, "%s", te->copyStmt);
+							AH->writingCopyData = true;
+						}
 
 						(*AH->PrintTocDataPtr) (AH, te, ropt);
 
+						AH->writingCopyData = false;
+
 						/*
 						 * If we just restored blobs, fix references in
 						 * previously-loaded tables; otherwise, if we
@@ -376,14 +383,12 @@ RestoreArchive(Archive *AHX, RestoreOptions *ropt)
 				_printTocEntry(AH, te, ropt, false, false);
 			}
 		}
-		te = te->next;
 	}							/* end loop over TOC entries */
 
 	/*
 	 * Scan TOC again to output ownership commands and ACLs
 	 */
-	te = AH->toc->next;
-	while (te != AH->toc)
+	for (te = AH->toc->next; te != AH->toc; te = te->next)
 	{
 		AH->currentTE = te;
 
@@ -396,10 +401,11 @@ RestoreArchive(Archive *AHX, RestoreOptions *ropt)
 				  te->desc, te->tag);
 			_printTocEntry(AH, te, ropt, false, true);
 		}
-
-		te = te->next;
 	}
 
+	if (AH->public.verbose)
+		dumpTimestamp(AH, "Completed on", time(NULL));
+
 	ahprintf(AH, "--\n-- PostgreSQL database dump complete\n--\n\n");
 
 	/*
@@ -890,24 +896,21 @@ SortTocFromFile(Archive *AHX, RestoreOptions *ropt)
 	if (!fh)
 		die_horribly(AH, modulename, "could not open TOC file\n");
 
-	while (fgets(buf, 1024, fh) != NULL)
+	while (fgets(buf, sizeof(buf), fh) != NULL)
 	{
-		/* Find a comment */
+		/* Truncate line at comment, if any */
 		cmnt = strchr(buf, ';');
-		if (cmnt == buf)
-			continue;
-
-		/* End string at comment */
 		if (cmnt != NULL)
 			cmnt[0] = '\0';
 
-		/* Skip if all spaces */
-		if (strspn(buf, " \t") == strlen(buf))
+		/* Ignore if all blank */
+		if (strspn(buf, " \t\r") == strlen(buf))
 			continue;
 
-		/* Get an ID */
+		/* Get an ID, check it's valid and not already seen */
 		id = strtol(buf, &endptr, 10);
-		if (endptr == buf || id <= 0 || id > AH->maxDumpId)
+		if (endptr == buf || id <= 0 || id > AH->maxDumpId ||
+			ropt->idWanted[id - 1])
 		{
 			write_msg(modulename, "WARNING: line ignored: %s\n", buf);
 			continue;
@@ -1275,7 +1278,8 @@ warn_or_die_horribly(ArchiveHandle *AH,
 	}
 	if (AH->currentTE != NULL && AH->currentTE != AH->lastErrorTE)
 	{
-		write_msg(modulename, "Error from TOC entry %d; %u %u %s %s %s\n", AH->currentTE->dumpId,
+		write_msg(modulename, "Error from TOC entry %d; %u %u %s %s %s\n",
+				  AH->currentTE->dumpId,
 		 AH->currentTE->catalogId.tableoid, AH->currentTE->catalogId.oid,
 		  AH->currentTE->desc, AH->currentTE->tag, AH->currentTE->owner);
 	}
@@ -1715,11 +1719,12 @@ _allocAH(const char *FileSpec, const ArchiveFormat fmt,
 
 	/*
 	 * On Windows, we need to use binary mode to read/write non-text archive
-	 * formats.  Force stdin/stdout into binary mode in case that is what
+	 * formats.  Force stdin/stdout into binary mode if that is what
 	 * we are using.
 	 */
 #ifdef WIN32
-	if (fmt != archNull)
+	if (fmt != archNull &&
+		(AH->fSpec == NULL || strcmp(AH->fSpec, "") == 0))
 	{
 		if (mode == archModeWrite)
 			setmode(fileno(stdout), O_BINARY);
@@ -2195,9 +2200,7 @@ _reconnectToDB(ArchiveHandle *AH, const char *dbname)
 
 		appendPQExpBuffer(qry, "\\connect %s\n\n",
 						  dbname ? fmtId(dbname) : "-");
-
-		ahprintf(AH, qry->data);
-
+		ahprintf(AH, "%s", qry->data);
 		destroyPQExpBuffer(qry);
 	}
 
@@ -2766,3 +2769,16 @@ checkSeek(FILE *fp)
 	else
 		return true;
 }
+
+
+/*
+ * dumpTimestamp
+ */
+static void
+dumpTimestamp(ArchiveHandle *AH, const char *msg, time_t tim)
+{
+	char		buf[256];
+
+	if (strftime(buf, 256, "%Y-%m-%d %H:%M:%S %Z", localtime(&tim)) != 0)
+		ahprintf(AH, "-- %s %s\n\n", msg, buf);
+}
